/*-----------------------------------------------------------------------

File  : can_infstate.c

Author: Stephan Schulz

Contents

  Functions for representing and analysing the state of an inference
  process.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Sun Feb 28 19:06:13 MET 1999
    New

-----------------------------------------------------------------------*/

#include "can_infstate.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

#ifndef NDEBUG

/*-----------------------------------------------------------------------
//
// Function: verify_clause_term_assoc()
//
//   For each term in clause check if term is associated with clause.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static bool verify_clause_term_assoc(InfState_p state, CompClause_p
				     clause, char* called)
{
   int i;
   PTree_p *root;
   Term_p dummy;
   Term_p term;

   assert(clause->lit_terms);
   
   for(i=0; i<2*clause->literal_no; i++)
   {
      term = clause->lit_terms[i];
      if(!TermIsVar(term))
      {
	 dummy = PDArrayElementP(state->term_clause_assoc,
				 term->entry_no);
	 root = (PTree_p*)
	    &(state->term_clause_assoc->array
	      [term->entry_no].p_val);
	 if(!PTreeFind(root, clause))
	 {
	    fprintf(GlobalOut, 
		    "# verify_clause_term_assoc(%s):"
		    " %ld is not associated with %ld\n",
		    called, clause->ident, term->entry_no);
	 }
      }
   }
   return true;
}
#endif

/*-----------------------------------------------------------------------
//
// Function: pcl_clause_output()
//
//   Print a PCL2 step from the provided data.
//
// Global Variables: -
//
// Side Effects    : Renames result clause
//
/----------------------------------------------------------------------*/

static void pcl_clause_output(FILE* out, InfState_p state,
			      CompClause_p concl, char* op,
			      CompClause_p main_premise, CompClause_p
			      side_premise, char *remark)
{
   long main_id=0, side_id=0;

   if(main_premise)
   {
      main_id = main_premise->ext_ident;
   }
   if(side_premise)
   {
      side_id = side_premise->ext_ident;
   }
   concl->ext_ident = state->ext_id_source++;
   
   fprintf(out, "%6ld : ", concl->ext_ident);
   CompClausePCLPrint(out, concl, state->terms);
   if(op!=PCL_QUOTE)
   {
      fprintf(out, " : %s", op);
      if(main_premise)
      {
	 fprintf(out, "(%ld", main_id);
	 if(side_premise)
	 {
	    fprintf(out, ",%ld",side_id);
	 }
	 fputc(')',out);
      }
      else
      {
	 assert(!side_premise);
      }
   }
   else
   {
      assert(main_premise);
      fprintf(out, " : %ld", main_id);
   }
   if(remark)
   {
      fprintf(out, " : {%s}", remark);
   }
}


/*-----------------------------------------------------------------------
//
// Function: pcl_step_conditional()
//
//   Print a PCL step if the output mode warants it.
//
// Global Variables: -
//
// Side Effects    : Output, may change identifiers
//
/----------------------------------------------------------------------*/

static void pcl_step_conditional(FILE* out, InfState_p state,
				 CompClause_p concl, InfType opcode,
				 CompClause_p main_premise,
				 CompClause_p side_premise,
				 char *remark)
{
   char* op = NULL;

   if(state->generate_pcl_all || (
      state->generate_pcl&&ClauseQueryProp(concl, CPIsWatched)))
   {
      switch(opcode)
      {
      case inf_initial:
	    op = "initial";
	    break;
      case inf_quote:
	    op = PCL_QUOTE;
	    break;
      case inf_paramod:
	    op = PCL_PM;
	    if(!side_premise)
	    {
	       side_premise = main_premise;
	    }
	    break;
      case inf_efactor:
	    op = PCL_EF;
	    break;
      case inf_eres:
	    op = PCL_ER;
	    break;
      case inf_simplify_reflect:
	    op = PCL_SR;
	    break;
      case inf_minimize:
	    op = PCL_CN;
	    break;
      case inf_rewrite:
	    op = PCL_RW;
	    assert(side_premise);
	    assert(side_premise!=main_premise);
	    break;
      case inf_mult_rewrite:
	    op = PCL_URW;
	    assert(side_premise);
	    assert(side_premise!=main_premise);
	    break;
      default:
	    assert(false);
	    break;
      }
      pcl_clause_output(out, state, concl, op, main_premise,
			side_premise, remark);
      fputc('\n',out);
   }
}


/*-----------------------------------------------------------------------
//
// Function: print_pcl_tree()
//
//   For all clauses in tree, conditionally print them as rewritten by
//   compclause. 
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void print_pcl_tree(FILE* out, PTree_p tree, CompClause_p
			   side_premise, InfState_p state)
{
   PStack_p stack;
   PTree_p  node;
   CompClause_p main_premise;

   stack = PTreeTraverseInit(tree);
   
   while((node = PTreeTraverseNext(stack)))
   {
      main_premise = node->key;
      pcl_step_conditional(out, state,  main_premise, inf_rewrite,
			   main_premise, side_premise, NULL);
   }
   PTreeTraverseExit(stack);
}


/*-----------------------------------------------------------------------
//
// Function: print_example_conditional()
//
//   Print a clause and appropriate information in the correct format
//   for an example file (if clause is watched).
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void print_example_conditional(FILE* out, CompClause_p clause,
				      InfState_p state)
{
   if(ClauseQueryProp(clause, CPIsWatched))
   {
      fprintf(out, "%6ld:", clause->ident);      
      if(state->print_clause_stats)
      {
	 ClauseStatsPrintNormalized(out, &(clause->stats),
				    state->created_count,
				    state->processed_count);
      }
      fprintf(out, ":");
      CompClausePrint(out, clause, state->terms, true);
      fprintf(out, "\n");
   }
}


/*-----------------------------------------------------------------------
//
// Function: print_example_tree()
//
//   For all clauses in tree, conditionally print them as examples. 
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void print_example_tree(FILE* out, PTree_p tree, InfState_p state)
{
   PStack_p stack;
   PTree_p  node;

   stack = PTreeTraverseInit(tree);
   
   while((node = PTreeTraverseNext(stack)))
   {
      print_example_conditional(out,node->key, state);
   }
   PTreeTraverseExit(stack);
}


/*-----------------------------------------------------------------------
//
// Function: find_associated_clauses()
//
//   Find all clauses associated with term and its superterms and add
//   them to the tree.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

static void find_associated_clauses(InfState_p state, Term_p term,
				    PTree_p *store)
{
   PStack_p ptree_stack, term_stack;
   PTree_p node;

   assert(term);
   
   term_stack = PStackAlloc();
   PStackPushP(term_stack, term);

   while(!PStackEmpty(term_stack))
   {
      term = PStackPopP(term_stack);
      assert(term);

      PTreeInsertTree(store, PDArrayElementP(state->term_clause_assoc,
					     term->entry_no));
      ptree_stack = PTreeTraverseInit(term->superterms);
      while((node = PTreeTraverseNext(ptree_stack)))
      {
         PStackPushP(term_stack, node->key);
      }  
      PTreeTraverseExit(ptree_stack);
   }
   PStackFree(term_stack);   
}


/*-----------------------------------------------------------------------
//
// Function: print_parent_tree()
//
//   Print the list of clause identifiers represented by parents.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void print_parent_tree(FILE* out, PTree_p parents)
{
   CompClause_p parent;
   PTree_p      cell;
   PStack_p     stack;
   char*        prefix = "";

   fprintf(out, "(");      
   stack = PTreeTraverseInit(parents);
   while((cell = PTreeTraverseNext(stack)))
   {
      parent = cell->key;
      assert(parent);
      fprintf(out, "%s%ld", prefix, parent->ident);
      prefix = ",";      
   }
   PTreeTraverseExit(stack);
   fprintf(out, ")");   
}


/*-----------------------------------------------------------------------
//
// Function: parse_parent_list()
//
//   Parse a list of parents for a clause and push them onto the
//   stack. Return number of arguments.
//
// Global Variables: -
//
// Side Effects    : Reads input
//
/----------------------------------------------------------------------*/

static long parse_parent_list(Scanner_p in, InfState_p state, PStack_p
			      parents)
{
   long         clause_id;
   CompClause_p compact; 
   long res = 0;

   assert(parents);

   AcceptInpTok(in, OpenBracket);
   
   while(!TestInpTok(in,CloseBracket))
   {
      CheckInpTok(in, PosInt);
      clause_id = AktToken(in)->numval;

      compact = PDArrayElementP(state->clause_store, clause_id);
      if(!compact)
      {
	 AktTokenError(in, "Unknown clause in parent list", false);
      }
      PStackPushP(parents,compact);
      assert(clause_id == compact->ident);
      NextToken(in);
      if(!TestInpTok(in,CloseBracket))
      {
	 AcceptInpTok(in, Comma);
      }
      res++;
   }
   AcceptInpTok(in, CloseBracket);
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: store_rewrite_parent()
//
//   For all clauses associated with term, add clause as a
//   rewrite-parent. 
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

static void store_rewrite_parent(InfState_p state, Term_p term,
				 CompClause_p clause)
{   
   PStack_p     clause_stack;
   CompClause_p handle;
   PTree_p      node;

   assert(clause);
   assert(clause->literal_no == 1);
   assert(state);
   assert(term);
   
   clause_stack =
      PTreeTraverseInit(PDArrayElementP(state->term_clause_assoc,
					term->entry_no));
   
   while((node = PTreeTraverseNext(clause_stack)))
   {
      handle = node->key;
      PTreeStore(&(handle->s_parents), clause);
   }
   PTreeTraverseExit(clause_stack);
}

/*-----------------------------------------------------------------------
//
// Function: store_rewrite_parents()
//
//   For all clauses associated with term and its superterms, add
//   clause as a rewrite-parent.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rek_store_rewrite_parent(InfState_p state, Term_p term,
				     CompClause_p clause)
{
   PStack_p term_stack, ptree_stack;
   PTree_p  node;

   assert(clause);
   assert(clause->literal_no == 1);
   assert(state);
   assert(term);

   term_stack = PStackAlloc();
   PStackPushP(term_stack, term);

   while(!PStackEmpty(term_stack))
   {
      term = PStackPopP(term_stack);
      assert(term);

      store_rewrite_parent(state, term, clause);
      
      ptree_stack = PTreeTraverseInit(term->superterms);
      while((node = PTreeTraverseNext(ptree_stack)))
      {
	 PStackPushP(term_stack, node->key);
      }	 
      PTreeTraverseExit(ptree_stack);
   }
   PStackFree(term_stack);   
}

/*-----------------------------------------------------------------------
//
// Function: change_term_clause_association()
//
//   Remove all clauses associated with term and associate them with
//   repl. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void change_term_clause_association(InfState_p state, Term_p
					   term, Term_p repl)
{
   PTree_p *root, source, *root1;
   long t_index, r_index;

   t_index = term->entry_no;
   r_index = repl->entry_no;

   assert(t_index >= 0);
   if(r_index >=0)
   {
      assert(t_index<=state->term_clause_assoc->size);
      /* assert(r_index<=state->term_clause_assoc->size); Wrong
	 assertion: repl has just been parsed, and thus does not need to
	 have any associated clauses. Moreover, in the prover repl has
	 just been generated by instanciation of a left hand side and
	 hence does not need to have any parents, either. */
      
      root1 = (PTree_p*)&PDArrayElementP(state->term_clause_assoc, r_index);
      
      root = (PTree_p*)
	 &(state->term_clause_assoc->array[r_index].p_val);
      assert(root == root1);
      source = state->term_clause_assoc->array[t_index].p_val;
      PTreeMerge(root, source);
   }
   else
   {
      source = state->term_clause_assoc->array[t_index].p_val;
      PTreeFree(source);
   } 
   PDArrayAssignP(state->term_clause_assoc, term->entry_no, NULL);
}


/*-----------------------------------------------------------------------
//
// Function: parse_term_list()
//
//   Parse a list of terms and push them on the stack.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static void parse_term_list(Scanner_p in, TB_p bank, PStack_p stack)
{
   Term_p term;
   
   AcceptInpTok(in, OpenSquare);

   while(!TestInpTok(in, CloseSquare))
   {
      term = TBTermParse(in, bank);
      PStackPushP(stack, term);
   }
   AcceptInpTok(in, CloseSquare); 
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: InfStateAlloc()
//
//   Allocate an empty inference process state.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

InfState_p InfStateAlloc()
{
   InfState_p handle = InfStateCellAlloc();

   handle->terms             = TBAlloc(TPRestricted,SigAlloc());
   handle->clause_store      = PDArrayAlloc(100000,100000);
   handle->term_clause_assoc = PDArrayAlloc(100000,100000);
   handle->watch_clauses     = NULL;
   
   handle->axiom_terms       = TBAlloc(TPIgnoreProps,
				       handle->terms->sig);   
   handle->axioms             = ClauseSetAlloc();
   handle->generate_pcl       = false;
   handle->generate_pcl_all   = false;
   handle->print_proc_pcl     = false;
   handle->print_clause_evals = false;
   handle->print_clause_mods  = false;
   handle->print_clause_stats = true;
   handle->update_inf_inc     = 1;
   handle->created_count      = 0;
   handle->processed_count    = 0;
   handle->ext_id_source      = 0;
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: InfStateFree()
//
//   Return all the memory taken up by a InfState.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void InfStateFree(InfState_p state)
{
   long i;
   CompClause_p clause;
   PTree_p tree;
   
   for(i=0; i<state->clause_store->size; i++)
   {
      clause = PDArrayElementP(state->clause_store,i);
      if(clause)
      {
	 CompClauseFree(clause, state->terms);
	 PDArrayAssignP(state->clause_store, i, NULL);
      }      
   }
   for(i=0; i<state->term_clause_assoc->size; i++)
   {
      tree = PDArrayElementP(state->term_clause_assoc,i);
      if(tree)
      {
	 PTreeFree(tree);
	 PDArrayAssignP(state->term_clause_assoc, i, NULL);
      }      
   }
   PDArrayFree(state->clause_store);
   PDArrayFree(state->term_clause_assoc);
   
   SigFree(state->terms->sig);
   state->terms->sig = NULL;
   TBFree(state->terms);
   ClauseStatTreeFree(state->watch_clauses);

   ClauseSetFree(state->axioms);
   state->axiom_terms->sig = NULL;
   TBFree(state->axiom_terms);

   InfStateCellFree(state);
}


/*-----------------------------------------------------------------------
//
// Function: InfStateClausesDelProp()
//
//   Delete the property from all clauses in state.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void InfStateClausesDelProp(InfState_p state, ClauseProperties props)
{
   long i;
   CompClause_p clause;
   
   for(i=0; i<state->clause_store->size; i++)
   {
      clause = PDArrayElementP(state->clause_store,i);
      if(clause)
      {
	 ClauseDelProp(clause, props);
      }      
   }   
}


/*-----------------------------------------------------------------------
//
// Function: InfStateClausesSetProp()
//
//   Set the property from all clauses in state.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void InfStateClausesSetProp(InfState_p state, ClauseProperties props)
{
   long i;
   CompClause_p clause;
   
   for(i=0; i<state->clause_store->size; i++)
   {
      clause = PDArrayElementP(state->clause_store,i);
      if(clause)
      {
	 ClauseSetProp(clause, props);
      }      
   }   
}


/*-----------------------------------------------------------------------
//
// Function: InfStateAddTermAssoc()
//
//   Given a state and a clause, add the term->clause associations to
//   the state.
//
// Global Variables: -
//
// Side Effects    : Changes state
//
/----------------------------------------------------------------------*/

void InfStateAddTermAssoc(InfState_p state, CompClause_p clause)
{
   int i;   
   PTree_p *root;
   Term_p dummy;

   assert(clause->lit_terms);
   
   for(i=0; i<2*clause->literal_no; i++)
   {
      if(!TermIsVar(clause->lit_terms[i]))
      {
	 /* Make sure that the dynamic array is large enough! */
	 dummy = PDArrayElementP(state->term_clause_assoc,
				 clause->lit_terms[i]->entry_no);
	 root = (PTree_p*)
	    &(state->term_clause_assoc->array
	      [clause->lit_terms[i]->entry_no].p_val);
	 PTreeStore(root, clause);
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: InfStateRemoveTermAssoc()
//
//   Given a state and a clause, remove the association of the terms
//   with the clause.
//
// Global Variables: - 
//
// Side Effects    : Changes state
//
/----------------------------------------------------------------------*/

void InfStateRemoveTermAssoc(InfState_p state, CompClause_p clause)
{
   int i;   
   PTree_p *root;
   Term_p dummy;

   assert(clause->lit_terms);
   
   for(i=0; i<2*clause->literal_no; i++)
   {
      if(!TermIsVar(clause->lit_terms[i]))
      {
	 dummy = PDArrayElementP(state->term_clause_assoc,
				 clause->lit_terms[i]->entry_no);
	 root = (PTree_p*)
	    &(state->term_clause_assoc->array
	      [clause->lit_terms[i]->entry_no].p_val);
	 PTreeDeleteEntry(root, clause);
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: InfStateCCParse()
//
//   Parse a CC (Clause creation) inference step into state.
//
// Global Variables: -
//
// Side Effects    : Changes state, reads input.
//
/----------------------------------------------------------------------*/

void InfStateCCParse(Scanner_p in, FILE* out, InfState_p state)
{
   Clause_p      clause;
   CompClause_p  compact;
   PStack_p      parents = PStackAlloc();
   long          clause_id;
   NumTree_p     cell;
   ClauseStats_p stats;
   InfType       just=inf_noinf;

   AcceptInpId(in, "CC");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   if(PDArrayElementP(state->clause_store, clause_id))
   {
      AktTokenError(in, "Clause to be created already exists", false);
   }
   /*  printf("CC %ld\n", clause_id); */
   NextToken(in);
   if(TestInpId(in, CC_AXIOM))
   {
      just=inf_initial;
   }
   else if(TestInpId(in, CC_PARAMOD))
   {
      just=inf_paramod;
   }
   else if(TestInpId(in, CC_EFACTOR))
   {
      just=inf_efactor;
   }
   else if(TestInpId(in, CC_OFACTOR))
   {
      just=inf_efactor; /* Ordered factor is special case of E-Factor! */
   }
   else if(TestInpId(in, CC_ERESOLUTION))
   {
      just=inf_eres;
   }
   AcceptInpId(in, CC_AXIOM "|" CC_PARAMOD "|" CC_EFACTOR "|"  
	           CC_OFACTOR "|" CC_ERESOLUTION);

   parse_parent_list(in, state, parents);
   clause = ClauseParse(in, state->terms);
   clause->ident = clause_id;

   if(just==inf_initial)
   {
      ClauseSetInsert(state->axioms,
		      ClauseCopy(clause, state->axiom_terms));
   }
		      
   compact = CompactifyClause(clause);
   assert(!compact->g_parents);
   PStackToPTree(&(compact->g_parents), parents);
   
   if((cell = NumTreeFind(&(state->watch_clauses), clause_id)))
   {      
      ClauseSetProp(compact, CPIsWatched);
      stats = cell->val1.p_val;
      compact->stats = *stats;
   }
   PDArrayAssignP(state->clause_store, clause_id, compact);   
   InfStateAddTermAssoc(state, compact);
   state->created_count += state->update_inf_inc;
   if(state->print_clause_mods)
   {
      print_example_conditional(out,compact, state);
   }
   PStackPushP(parents, NULL); /* Ensure that there are at least two */
   PStackPushP(parents, NULL); /* valid parents */
   pcl_step_conditional(out, state, compact, just,
			PStackElementP(parents,0),
			PStackElementP(parents,1), 
			NULL); 
   PStackFree(parents);
   assert(verify_clause_term_assoc(state, compact, "CC"));
}


/*-----------------------------------------------------------------------
//
// Function: InfStateCRParse()
//
//   Parse a CR (Clause renaming) inference step into state.
//
// Global Variables: -
//
// Side Effects    : Changes state, reads input.
//
/----------------------------------------------------------------------*/

void InfStateCRParse(Scanner_p in, FILE* out, InfState_p state)
{
   Clause_p clause;
   CompClause_p compact;
   PStack_p     parents = PStackAlloc();
   long         clause_id;
   long         parent_no;
   CompClause_p parent;
   NumTree_p     cell;
   ClauseStats_p stats;

   AcceptInpId(in, "CR");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   if(PDArrayElementP(state->clause_store, clause_id))
   {
      AktTokenError(in, "Target clause of renaming already exists", false);
   }
   /*  printf("CR %ld\n", clause_id); */
   NextToken(in);
   AcceptInpId(in, "RENAME");
   parent_no = parse_parent_list(in, state, parents);
   if(parent_no != 1)
   {
      AktTokenError(in,
		    "RENAME expression takes exactly one argument",
		    false);
   }
   parent = PStackElementP(parents,0);
   
   clause = UnpackClause(parent, state->terms);
   clause->ident = clause_id;    
   compact = CompactifyClause(clause);
   compact->ext_ident = parent->ext_ident;
   assert(!compact->g_parents);
   PStackToPTree(&(compact->g_parents), parents);
   if((cell = NumTreeFind(&(state->watch_clauses), clause_id)))
   {      
      ClauseSetProp(compact, CPIsWatched);
      stats = cell->val1.p_val;
      compact->stats = *stats;
   }
   PDArrayAssignP(state->clause_store, clause_id, compact);   
   InfStateAddTermAssoc(state, compact);
   
   ClauseSetProp(parent, CPDeleteClause);
   InfStateRemoveTermAssoc(state, parent);
   CompClauseRemoveTerms(parent, state->terms);
   PStackFree(parents);
   assert(verify_clause_term_assoc(state, compact, "CR"));
}


/*-----------------------------------------------------------------------
//
// Function: InfStateSSParse()
//
//   Parse a SS (subsumption) inference step into state.
//
// Global Variables: -
//
// Side Effects    : Changes state, reads input.
//
/----------------------------------------------------------------------*/

void InfStateSSParse(Scanner_p in, FILE* out, InfState_p state)
{
   long         clause_id;
   long         subsumer_id;
   CompClause_p subsumer;

   AcceptInpId(in, "SS");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   if(!PDArrayElementP(state->clause_store, clause_id))
   {
      AktTokenError(in, "Subsumed clause in SS inference does not exist", false);
   }
   NextToken(in);
   AcceptInpId(in, CS_SUBSUME);
   AcceptInpTok(in, OpenBracket);
   CheckInpTok(in, PosInt);   
   subsumer_id = AktToken(in)->numval;
   subsumer = PDArrayElementP(state->clause_store, subsumer_id);
   if(!subsumer)
   {
      AktTokenError(in, "Subsuming clause in SS inference does not exist", false);
   }
   NextToken(in);   
   AcceptInpTok(in, CloseBracket);
   subsumer->stats.subsumed++;
}


/*-----------------------------------------------------------------------
//
// Function: InfStateCPParse()
//
//   Parse a CP (Clause processing) inference.
//
// Global Variables: -
//
// Side Effects    : Adds some properties to clauses.
//
/----------------------------------------------------------------------*/

void InfStateCPParse(Scanner_p in, FILE* out, InfState_p state)
{
   long clause_id;
   CompClause_p compact;

   AcceptInpId(in, "CP");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause selected for processing",
		    false);
   }
   NextToken(in);
   assert(clause_id == compact->ident);
   CheckInpId(in, CP_PROCESS "|" CP_PROOF "|" CP_FINAL "|" CP_EXISTS);
   if(TestInpId(in, CP_PROCESS))
   {
      ClauseSetProp(compact, CPIsSelected);
      if(state->print_proc_pcl)
      {
	 pcl_step_conditional(out, state, compact, inf_quote, compact,
			      NULL, "proc"); 
      }
   }
   else if(TestInpId(in, CP_PROOF))
   {
      ClauseSetProp(compact, CPIsFinal|CPIsSelected);
      pcl_step_conditional(out, state, compact, inf_quote, compact,
			   NULL, "proof");
   }
   else if(TestInpId(in, CP_FINAL))
   {
      ClauseSetProp(compact, CPIsFinal);
      pcl_step_conditional(out, state, compact, inf_quote, compact,
			   NULL, "final");
   }
   else
   {
      assert(TestInpId(in, CP_EXISTS));
      if(state->print_proc_pcl)
      {
	 pcl_step_conditional(out, state, compact, inf_quote, compact,
			      NULL, NULL); 
      }
   }
   NextToken(in);
   state->processed_count += state->update_inf_inc;
   assert(verify_clause_term_assoc(state, compact, "CP"));
}

/*-----------------------------------------------------------------------
//
// Function: InfStateCSParse()
//
//   Parse a CS (Clause status change) inference.
//
// Global Variables: -
//
// Side Effects    : May change status of clause.
//
/----------------------------------------------------------------------*/

void InfStateCSParse(Scanner_p in, FILE* out, InfState_p state)
{
   long clause_id;
   CompClause_p compact;

   AcceptInpId(in, "CS");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause specified for status change",
		    false);
   }
   assert(verify_clause_term_assoc(state, compact, "CS"));

   NextToken(in);
   assert(clause_id == compact->ident);
   
   AcceptInpId(in, CS_DELETE);
   ClauseSetProp(compact, CPDeleteClause);
   InfStateRemoveTermAssoc(state, compact);
   CompClauseRemoveTerms(compact, state->terms);
}

/*-----------------------------------------------------------------------
//
// Function: InfStateCMParse()
//
//   Parse a CM (Clause Modify) inference.
//
// Global Variables: -
//
// Side Effects    : Changes term associations
//
/----------------------------------------------------------------------*/

void InfStateCMParse(Scanner_p in, FILE* out, InfState_p state)
{
   Clause_p     clause;
   CompClause_p compact;
   long         clause_id;
   bool         ext_rep_change = false;
   PStack_p     parents = PStackAlloc();
   InfType       just=inf_noinf;

   AcceptInpId(in, "CM");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause specified for modification",
		    false);
   }
   NextToken(in);
   /* printf("CM %ld\n", clause_id); */
   assert(clause_id == compact->ident);
   
   if(TestInpId(in, CM_SIMPLIFY_REFLECT))
   {
      ext_rep_change = true;
      just = inf_simplify_reflect;
   }
   else if(TestInpId(in, CM_MINIMIZE))
   {
      ext_rep_change = true;
      just = inf_minimize;
   }
   AcceptInpId(in, CM_SIMPLIFY_REFLECT "|" CM_TERM_RENAME "|" CM_MINIMIZE);

   parse_parent_list(in, state, parents);
   PStackToPTree(&(compact->g_parents), parents);

   clause = ClauseParse(in, state->terms);
   InfStateRemoveTermAssoc(state, compact);
   CompClauseRemoveTerms(compact, state->terms);
   CompClauseAddTerms(compact, clause);
   InfStateAddTermAssoc(state, compact);   
   ClauseFree(clause);
   if(state->print_clause_mods && ext_rep_change)
   {
      print_example_conditional(out,compact, state);
   }
   if(just==inf_simplify_reflect)
   {
       pcl_step_conditional(out, state, compact, just, compact,
			    PStackElementP(parents,0), NULL);
   }
   else if(just==inf_minimize)
   {
      pcl_step_conditional(out, state, compact, just, compact, NULL,
			   NULL);
   }
   PStackFree(parents);
   assert(verify_clause_term_assoc(state, compact, "CM"));
}


/*-----------------------------------------------------------------------
//
// Function: InfStateCEParse()
//
//   Parse a CE (Clause Evaluate) inference.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void InfStateCEParse(Scanner_p in, FILE* out, InfState_p state)
{
   CompClause_p compact;
   long         clause_id;
   
   AcceptInpId(in, "CE");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause for evaluation",
		    false);
   }
   NextToken(in);
   assert(clause_id == compact->ident);
   
   if(state->print_clause_evals)
   {
      print_example_conditional(out, compact, state);
      pcl_step_conditional(out, state, compact, inf_quote, compact,
			   NULL, "eval");
   }
   assert(verify_clause_term_assoc(state, compact, "CE"));
}


/*-----------------------------------------------------------------------
//
// Function: InfStateRWParse()
//
//   Parse a RW (rewrite) inference into state.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void InfStateRWParse(Scanner_p in, FILE* out, InfState_p state)
{
   long clause_id;
   Term_p term, repl, t1, r1;
   CompClause_p compact;
   PStack_p super_stack = PStackAlloc(), tmp_stack;
   PTree_p  affected_clauses = NULL;

   AcceptInpId(in, "RW");
   CheckInpTok(in, PosInt);   
   clause_id = AktToken(in)->numval;
   compact = PDArrayElementP(state->clause_store, clause_id);
   if(!compact)
   {
      AktTokenError(in, "Unknown clause specified as rewrite-rule",
		    false);
   }
   NextToken(in);
   assert(clause_id == compact->ident);
   
   term = TBTermParse(in, state->terms);
   AcceptInpTok(in, LesserSign);
   AcceptInpTokNoSkip(in, EqualSign);
   repl = TBTermParse(in, state->terms);

   if(state->print_clause_mods||
      state->generate_pcl||
      state->generate_pcl_all)
   {
      find_associated_clauses(state, term, &affected_clauses);
   }
   
   tmp_stack = PStackAlloc();
   PStackPushP(tmp_stack, term);
   PStackPushP(tmp_stack, repl);
   parse_term_list(in, state->terms, tmp_stack);
   PStackPushP(super_stack, tmp_stack);

   rek_store_rewrite_parent(state, term, compact);
   change_term_clause_association(state, term, repl);
   while(!TestInpTok(in, Fullstop))
   {
      t1 = TBTermParse(in, state->terms);
      AcceptInpTok(in, LesserSign);
      AcceptInpTokNoSkip(in, EqualSign);
      r1 = TBTermParse(in, state->terms);
      tmp_stack = PStackAlloc();
      PStackPushP(tmp_stack, t1);
      PStackPushP(tmp_stack, r1);
      parse_term_list(in, state->terms, tmp_stack);
      PStackPushP(super_stack, tmp_stack);
      change_term_clause_association(state, t1, r1);
   }
   AcceptInpTok(in, Fullstop);
   TBTermReplaceRepro(state->terms, term, repl, super_stack);
   
   while(!PStackEmpty(super_stack))
   {
      PStackFree(PStackPopP(super_stack));
   }
   PStackFree(super_stack);

   if(state->print_clause_mods)
   {
      print_example_tree(out, affected_clauses, state);
   }   
   if(state->generate_pcl||
      state->generate_pcl_all)
   {
      print_pcl_tree(out, affected_clauses, compact, state);
   }
   if(state->print_clause_mods||
      state->generate_pcl||
      state->generate_pcl_all)
   {
      PTreeFree(affected_clauses);
   }
}



/*-----------------------------------------------------------------------
//
// Function: InfStateInferenceParse() 
//
//   Parse an arbitrary inference into state.
//
// Global Variables: -
//
// Side Effects    : Reads input, changes state
//
/----------------------------------------------------------------------*/

void InfStateInferenceParse(Scanner_p in, FILE* out, InfState_p state)
{
   /* printf("Parsing line %ld\n", AktToken(in)->line); */
   if(TestInpId(in, "CC"))
   {
      InfStateCCParse(in, out, state);
   }
   else if(TestInpId(in, "SS"))
   {
      InfStateSSParse(in, out, state);
   }
   else if(TestInpId(in, "CP"))
   {
      InfStateCPParse(in, out, state);
   }
   else if(TestInpId(in, "CS"))
   {
      InfStateCSParse(in, out, state);
   }
   else if(TestInpId(in, "CM"))
   {
      InfStateCMParse(in, out, state);
   }
   else if(TestInpId(in, "CR"))
   {
      InfStateCRParse(in, out, state);
   }
   else if(TestInpId(in, "CE"))
   {
      InfStateCEParse(in, out, state);
   }
   else if(TestInpId(in, "RW"))
   {
      InfStateRWParse(in, out, state);
   }
   else
   {
      AcceptInpId(in, "CC|SS|CP|CM|RW|CS|CR|CE");
      assert(false && "Not yet implemented");
   }
}


/*-----------------------------------------------------------------------
//
// Function: InfStateInfListParse()
//
//   Parse a list of inference step descriptions into the
//   state. Return number of steps parsed.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long InfStateInfListParse(Scanner_p in, FILE* out, InfState_p state)
{
   long steps = 0;

   while(!TestInpTok(in,NoToken))
   {
      InfStateInferenceParse(in, out, state);
      steps++;
   }
   return steps;
}


/*-----------------------------------------------------------------------
//
// Function: InfStatePrintClauses()
//
//   Print the selected clauses and information about them.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void InfStatePrintClauses(FILE* out, InfState_p state,
			  ClauseProperties filter, bool status, bool
			  deps, bool stats)
{
   long i;
   CompClause_p clause;

   for(i=0; i<state->clause_store->size; i++)
   {
      clause = PDArrayElementP(state->clause_store,i);
      if(clause && ClauseQueryProp(clause, filter))
      {
	 fprintf(out, "%ld", i);
	 if(status)
	 {
	    fprintf(out, ":%c,%c",
		    ClauseQueryProp(clause, CPIsSelected)?'S':'N',
		    ClauseQueryProp(clause, CPIsProofClause)?'P':'N');
	 }
	 if(deps)
	 {
	    fprintf(out, ":");
	    print_parent_tree(out, clause->g_parents);
	    print_parent_tree(out, clause->s_parents);	 
	 }
	 if(stats)
	 {
	    fprintf(out, ":");
	    ClauseStatsPrint(out, &(clause->stats));
	 }
	 fprintf(out, "\n");
      }
   }
}

/*-----------------------------------------------------------------------
//
// Function: InfStateStoreClauseInfo()
//
//   Store the clausestate-Information about selected clauses 
//   in the tree "store".
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

long InfStateStoreClauseInfo(NumTree_p *store, InfState_p state)
{
   long i, count = 0;
   CompClause_p clause;
   long key;
   IntOrP tmp;
   ClauseStats_p stats;
   bool control;

   for(i=0; i<state->clause_store->size; i++)
   {
      clause = PDArrayElementP(state->clause_store,i);
      if(clause && ClauseQueryProp(clause, CPOpFlag))
      {
	 count++;
	 key = i;
	 stats = ClauseStatsCopy(&(clause->stats));
	 tmp.p_val = stats;
	 control = NumTreeStore(store, key, tmp, tmp);
	 assert(control);
      }
   }
   return count;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/













