/*-----------------------------------------------------------------------

File  : ccl_clausesets.c

Author: Stephan Schulz

Contents
 
  Implementation of clause sets based on AVL trees.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Sun May 10 03:03:20 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_clausesets.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: print_var_pattern()
//
//   Print a template for a function/predicate symbol.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void print_var_pattern(FILE* out, char* symbol, int arity, char*
			      var, char* alt_var, int exception)
{
   int i;
   char* prefix = "";

   fprintf(out, "%s(", symbol);
   
   for(i=1; i<= arity; i++)
   {
      fprintf(out, prefix);
      if(i==exception)
      {
	 fprintf(out, alt_var);
      }
      else
      {
	 fprintf(out, "%s%d", var, i);
      }
      prefix = ",";
   }
   fprintf(out, ")");
}


/*-----------------------------------------------------------------------
//
// Function:  eq_func_axiom_print()
//
//   Print the LOP substitutivity axiom(s) for a function symbol.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void eq_func_axiom_print(FILE* out, char* symbol, int arity,
				bool single_subst)
{
   int i;
   char *prefix = "";
   
   if(single_subst)
   {
      for(i=1; i<=arity; i++)
      {
	 fprintf(out, "equal(");
	 print_var_pattern(out, symbol, arity, "X", "Y", i);
	 fprintf(out, ",");
	 print_var_pattern(out, symbol, arity, "X", "Z", i);	 
	 fprintf(out, ") <- ");
	 fprintf(out, "equal(Y,Z).\n");
      }
   }
   else
   {
      fprintf(out, "equal(");
      print_var_pattern(out, symbol, arity, "X", NULL,0);
      fprintf(out, ",");
      print_var_pattern(out, symbol, arity, "Y", NULL,0);
      fprintf(out, ") <- ");
      for(i=1; i<=arity; i++)
      {
	 fprintf(out, "%sequal(X%d,Y%d)", prefix, i, i);
	 prefix = ",";
      }
      fprintf(out, ".\n");
   }
}

/*-----------------------------------------------------------------------
//
// Function:  eq_pred_axiom_print()
//
//   Print the LOP substitutivity axiom(s) for a predicate symbol.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void eq_pred_axiom_print(FILE* out, char* symbol, int arity,
				bool single_subst)
{
   int i;
   
   if(single_subst)
   {
      for(i=1; i<=arity; i++)
      {
	 print_var_pattern(out, symbol, arity, "X", "Y", i);
	 fprintf(out, " <- ");
	 print_var_pattern(out, symbol, arity, "X", "Z", i);
	 fprintf(out, ", equal(Y,Z).\n");
      }
   }
   else
   {
      print_var_pattern(out, symbol, arity, "X", NULL,0);
      fprintf(out, " <- ");
      print_var_pattern(out, symbol, arity, "Y", NULL,0);
      for(i=1; i<=arity; i++)
      {
	 fprintf(out, ",equal(X%d,Y%d)", i, i);
      }
      fprintf(out, ".\n");
   }
}


/*-----------------------------------------------------------------------
//
// Function:  tptp_eq_func_axiom_print()
//
//   Print the TPTP substitutivity axiom(s) for a function symbol.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void tptp_eq_func_axiom_print(FILE* out, char* symbol, int arity,
				     bool single_subst)
{
   int i;
   
   if(single_subst)
   {
      for(i=1; i<=arity; i++)
      {	 
	 fprintf(out, "input_clause(eq_subst_%s%d, axiom, [++equal(",
		 symbol, i);
	 print_var_pattern(out, symbol, arity, "X", "Y", i);
	 fprintf(out, ",");
	 print_var_pattern(out, symbol, arity, "X", "Z", i);	 
	 fprintf(out, "),");
	 fprintf(out, "--equal(Y,Z)]).\n");
      }
   }
   else
   {
      fprintf(out, "input_clause(eq_subst_%s, axiom, [++equal(",
	      symbol);
      print_var_pattern(out, symbol, arity, "X", NULL,0);
      fprintf(out, ",");
      print_var_pattern(out, symbol, arity, "Y", NULL,0);
      fprintf(out, ")");
      for(i=1; i<=arity; i++)
      {
	 fprintf(out, ",--equal(X%d,Y%d)", i, i);
      }
      fprintf(out, "]).\n");
   }
}

/*-----------------------------------------------------------------------
//
// Function:  tptp_eq_pred_axiom_print()
//
//   Print the TPTP substitutivity axiom(s) for a predicate symbol.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void tptp_eq_pred_axiom_print(FILE* out, char* symbol, int arity,
				     bool single_subst)
{
   int i;
   
   if(single_subst)
   {
      for(i=1; i<=arity; i++)
      {
	 fprintf(out, "input_clause(eq_subst_%s%d, axiom, [++",
		 symbol, i);
	 print_var_pattern(out, symbol, arity, "X", "Y", i);
	 fprintf(out, ",--");
	 print_var_pattern(out, symbol, arity, "X", "Z", i);	 
	 fprintf(out, ",--equal(Y,Z)]).\n");
      }
   }
   else
   {
      fprintf(out, "input_clause(eq_subst_%s, axiom, [++",
	      symbol);
      print_var_pattern(out, symbol, arity, "X", NULL,0);
      fprintf(out, ",--");
      print_var_pattern(out, symbol, arity, "Y", NULL,0);
      for(i=1; i<=arity; i++)
      {
	 fprintf(out, ",--equal(X%d,Y%d)", i, i);
      }
      fprintf(out, "]).\n");
   }
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ClauseSetAlloc()
//
//   Allocate an empty clause set that uses setclock for time-keeping.
//
// Global Variables: -
//
// Side Effects    : Memory management
//
/----------------------------------------------------------------------*/

ClauseSet_p ClauseSetAlloc()
{
   ClauseSet_p handle;

   handle = ClauseSetCellAlloc();
   
   handle->members = 0;
   handle->literals = 0;
   handle->anchor = ClauseCellAlloc();
   handle->anchor->literals = NULL;
   handle->anchor->pred = handle->anchor->succ = handle->anchor;
   handle->date = SysDateCreationTime();
   SysDateInc(&handle->date);
   handle->demod_index = NULL;
   
   handle->eval_indices = PDArrayAlloc(4,4);

   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseSetFree()
//
//   Delete a clauseset.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseSetFree(ClauseSet_p junk)
{
   Clause_p handle;
   
   assert(junk);

   handle = junk->anchor->succ;
   while(handle!=junk->anchor)
   {
      handle = handle->succ;
      handle->pred->set = NULL;
      ClauseFree(handle->pred);
   }
   if(junk->demod_index)
   {
      PDTreeFree(junk->demod_index);
   }
   PDArrayFree(junk->eval_indices);
   ClauseCellFree(junk->anchor);
   ClauseSetCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetInsert() 
//
//   Insert a clause as the last clause into the clauseset.
//
// Global Variables: -
//
// Side Effects    : Changes set
//
/----------------------------------------------------------------------*/

void ClauseSetInsert(ClauseSet_p set, Clause_p new)
{
   int    i;
   Eval_p handle, test;

   assert(!new->set);

   new->succ = set->anchor;
   new->pred = set->anchor->pred;
   set->anchor->pred->succ = new;
   set->anchor->pred = new;
   new->set = set;
   set->members++;
   set->literals+=ClauseLiteralNumber(new);

   i=0;
   for(handle = new->evaluations; handle; handle = handle->next)
   {
      test = EvalTreeInsert((Eval_p*)
			    PDArrayElementRef(set->eval_indices, i), handle);
      assert(!test);
      i++;
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetIndexedInsert()
//
//   Insert a demodulator into the set and the sets index.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void ClauseSetIndexedInsert(ClauseSet_p set, Clause_p new)
{
   ClausePos_p pos;

   assert(set->demod_index);
   assert(ClauseIsDemodulator(new));
   
   ClauseSetInsert(set, new);
   pos = ClausePosAlloc();
   pos->clause = new;
   pos->literal = new->literals;
   pos->side = LeftSide;
   PDTreeInsert(set->demod_index, pos);
   if(!EqnIsOriented(new->literals))
   {
      pos = ClausePosAlloc();
      pos->clause = new;
      pos->literal = new->literals;
      pos->side = RightSide;
      PDTreeInsert(set->demod_index, pos);
   }
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetExtractEntry()
//
//   Remove a clause from a clause set.
//
// Global Variables: -
//
// Side Effects    : Changes set
//
/----------------------------------------------------------------------*/

void ClauseSetExtractEntry(Clause_p clause)
{
   int     i;
   Eval_p  handle, test;

   assert(clause->set);
   assert(clause->set->members);
   /* assert(ClauseSetFind(clause->set, clause));  */
   
   /* printf("ClauseSetExtractEntry: %d\n", (int)clause); */

   i=0;
   for(handle = clause->evaluations; handle; handle = handle->next)
   {
      /* EvalListPrint(stdout, handle);printf("\n"); */
      test = EvalTreeExtractEntry((Eval_p*)
	 PDArrayElementRef(clause->set->eval_indices, i), handle); 
      assert(test);
      assert(test->object == clause);
      i++;
   }
   clause->pred->succ = clause->succ;
   clause->succ->pred = clause->pred;
   clause->set->literals-=ClauseLiteralNumber(clause);
   clause->set->members--;
   clause->set = NULL;
   clause->succ = NULL;
   clause->pred = NULL;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetIndexedExtractEntry()
//
//   Remove a clause from an indexed clause set.
//
// Global Variables: -
//
// Side Effects    : Changes set
//
/----------------------------------------------------------------------*/

void ClauseSetIndexedExtractEntry(Clause_p clause)
{
   if(clause->set->demod_index)
   {
      assert(ClauseIsDemodulator(clause));
      /* printf("ClauseSetIndexedExtractEntry() left\n"); */
      PDTreeDelete(clause->set->demod_index, clause->literals->lterm,
		   clause); 
      if(!EqnIsOriented(clause->literals))
      {
	 /* printf("ClauseSetIndexedExtractEntry() right\n"); */
	 PDTreeDelete(clause->set->demod_index,
		      clause->literals->rterm, clause);
      }
   }
   ClauseSetExtractEntry(clause);
}  


/*-----------------------------------------------------------------------
//
// Function: ClauseSetExtractFirst()
//
//   Extract the first element of the set and return it. Return NULL
//   if set is empty.
//
// Global Variables: -
//
// Side Effects    : As above
//
/----------------------------------------------------------------------*/

Clause_p ClauseSetExtractFirst(ClauseSet_p set)
{
   Clause_p handle;

   if(ClauseSetEmpty(set))
   {
      return NULL;
   }
   handle = set->anchor->succ;
   assert(handle->set == set);
   ClauseSetExtractEntry(handle);
   
   return handle;
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetDeleteEntry()
//
//   Delete a clause from the clause set. 
//
// Global Variables: -
//
// Side Effects    : Changes tree, memory operations
//
/----------------------------------------------------------------------*/

void ClauseSetDeleteEntry(Clause_p clause)
{
   ClauseDefaultDocumentStatusChange(clause, CS_DELETE);
   ClauseSetExtractEntry(clause);
   ClauseFree(clause);
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetIndexedDeleteEntry()
//
//   Delete a clause from the clause set. 
//
// Global Variables: -
//
// Side Effects    : Changes tree, memory operations
//
/----------------------------------------------------------------------*/

void ClauseSetIndexedDeleteEntry(Clause_p clause)
{
   ClauseDefaultDocumentStatusChange(clause, CS_DELETE);
   ClauseSetIndexedExtractEntry(clause);
   ClauseFree(clause);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetFindBest()
//
//   Find the best clause (i.e. the clause with the smallest
//   evaluation).
//
// Global Variables: -
//
// Side Effects    : Changes clause set
//
/----------------------------------------------------------------------*/

Clause_p ClauseSetFindBest(ClauseSet_p set, int idx)
{
   Clause_p clause;
   Eval_p   evaluation;

   evaluation =
      EvalTreeFindSmallest(PDArrayElementP(set->eval_indices, idx));
   if(!evaluation)
   {
      assert(set->anchor->succ == set->anchor);
      return NULL;
   }
   assert(evaluation->object);
   clause = evaluation->object;
   assert(clause->set == set);
   return clause;
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetPrint()
//
//   Print the clause set to the given stream.
//
// Global Variables: Only for term output
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClauseSetPrint(FILE* out, ClauseSet_p set, bool fullterms)
{
   Clause_p handle;

   for(handle = set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      ClausePrint(out, handle, fullterms);
      fputc('\n', out);
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetSetProp()
//
//   Set prop in all clauses in set.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void ClauseSetSetProp(ClauseSet_p set, ClauseProperties prop)
{
   Clause_p handle;

   for(handle = set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      ClauseSetProp(handle, prop);
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetDelProp()
//
//   Delete prop in all clauses in set.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void ClauseSetDelProp(ClauseSet_p set, ClauseProperties prop)
{
   Clause_p handle;

   for(handle = set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      ClauseDelProp(handle, prop);
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetMarkCopies()
//
//   Mark clauses that are equivalent (modulo ClauseCompareFun) to
//   clauses that occur earlier in set. Returns number of marked
//   clauses. 
//
// Global Variables: -
//
// Side Effects    : Sets property, memory operations, deletes parents
//                   of unmarked copyes.
//
/----------------------------------------------------------------------*/

long ClauseSetMarkCopies(ClauseSet_p set)
{
   long res = 0;
   Clause_p handle, exists;
   PTree_p  store = NULL;

   assert(set);

   for(handle = set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      if((exists = PTreeObjStore(&store, handle,
			       (ComparisonFunctionType)ClauseCompareFun)))
      {
	 if(!ClauseParentsAreSubset(exists, handle))
	 {
	    ClauseDetachParents(exists);
	 }
	 ClauseSetProp(handle, CPDeleteClause);
	 res++;
      }
   }
   PTreeFree(store);
   
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetDeleteMarkedEntries()
//
//   Remove all clauses with property CPDeleteClause set. Returns
//   number of deleted clauses.
//
// Global Variables: -
//
// Side Effects    : Changes set.
//
/----------------------------------------------------------------------*/

long ClauseSetDeleteMarkedEntries(ClauseSet_p set)
{
   long deleted = 0;
   Clause_p clause, handle;
   
   assert(set);

   handle = set->anchor->succ;
   
   while(handle != set->anchor)
   {
      clause = handle;
      handle = handle->succ;

      if(ClauseQueryProp(clause, CPDeleteClause))
      {
	 deleted++;
	 ClauseDetachParents(clause); /* If no parents, nothing should
					 happen! */
	 ClauseSetDeleteEntry(clause);
      }
   }
   return deleted;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetDeleteCopies()
//
//   Delete all but one occurence of a clause in set.
//
// Global Variables: -
//
// Side Effects    : Changes clause set, changes parent-rlation of
//                   kept copies.
//
/----------------------------------------------------------------------*/

long ClauseSetDeleteCopies(ClauseSet_p set)
{
   long res1, res2;

   assert(set);

   res1 = ClauseSetMarkCopies(set);
   res2 = ClauseSetDeleteMarkedEntries(set);
   assert(res1==res2);

   return res1;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetDeleteNonUnits()
//
//   Remove all non-empty-non-unit-clauses from set, return number of
//   clauses eleminated. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long ClauseSetDeleteNonUnits(ClauseSet_p set)
{
   Clause_p handle;

   assert(set);
   assert(!set->demod_index);
   
   handle = set->anchor->succ;
   while(handle != set->anchor)
   {      
      if(ClauseLiteralNumber(handle)>1)
      {
	 ClauseSetProp(handle,CPDeleteClause);
      }
      else
      {
	 ClauseDelProp(handle,CPDeleteClause);	 
      }
      handle = handle->succ;
   }
   return ClauseSetDeleteMarkedEntries(set);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetGetTermNodes()
//
//   Count the nodes of terms in the clauses of set as though they
//   were unshared.
//
// Global Variables: -
//
// Side Effects    : Potentially via ClauseWeight()
//
/----------------------------------------------------------------------*/

long ClauseSetGetTermNodes(ClauseSet_p set)
{
   long     res = 0;
   Clause_p handle;
   
   for(handle = set->anchor->succ; handle != set->anchor; handle =
	  handle->succ)
   {
      res += ClauseWeight(handle, 1, 1, 1, 1, 1, true);
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetParseList()
//
//   Parse a list of clauses into the set. Clauses are not
//   evaluated. Returns number of clauses parsed.
//
// Global Variables: Only for parsing of terms.
//
// Side Effects    : Input, memory operations, changes set.
//
/----------------------------------------------------------------------*/

long ClauseSetParseList(Scanner_p in, ClauseSet_p set, TB_p bank)
{
   long     count = 0;
   Clause_p clause;
   
   while(ClauseStartsMaybe(in))
   {
      clause = ClauseParse(in, bank);
      count++;
      ClauseSetInsert(set, clause);
   }
   return count;
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetMarkMaximalTerms()
//
//   Orient all literals and mark all maximal terms and literals in
//   the set.
//
// Global Variables: -
//
// Side Effects    : Orients and marks literals
//
/----------------------------------------------------------------------*/

void ClauseSetMarkMaximalTerms(OCB_p ocb, ClauseSet_p set)
{
   Clause_p handle;

   for(handle = set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      ClauseMarkMaximalTerms(ocb, handle);
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseKillChildren()
//
//   Remove and delete children of a clause.
//
// Global Variables: -
//
// Side Effects    : Changes term sets
//
/----------------------------------------------------------------------*/

void ClauseKillChildren(Clause_p clause)
{
   Clause_p current;

   while(clause->children)
   {
      current = clause->children->key;
      assert(clause == current->parent1 || clause == current->parent2);
      /* printf("ClauseKillChildren(%d) -> %d\n",
	 (int)clause,(int)current); */
      ClauseDetachParents(current);
      if(current->set->demod_index)
      {
	 ClauseSetIndexedDeleteEntry(current);
      }
      else
      {
	 ClauseSetDeleteEntry(current);
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveSuperfluousLiterals()
//
//   Remove duplicate and trivial negative literals from the
//   clause. Return number of removed literals.
//
// Global Variables: -
//
// Side Effects    : Changes clause, termbanks
//
/----------------------------------------------------------------------*/

int ClauseRemoveSuperfluousLiterals(Clause_p clause)
{
   Eqn_p handle;
   int   removed = 0;

   removed += EqnListRemoveResolved(&(clause->literals), TBTermEqual);
   removed += EqnListRemoveDuplicates(clause->literals, TBTermEqual);
   
   clause->neg_lit_no = 0;
   clause->pos_lit_no = 0;   
   handle = clause->literals;

   while(handle)
   {
      if(EqnIsPositive(handle))
      {
	 clause->pos_lit_no++;
      }
      else
      {
	 clause->neg_lit_no++;
      }
      handle = handle->next;
   }
   
   if(clause->set)
   {
      clause->set->literals-=removed;
   }
   return removed;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetListGetMaxDate()
//
//   Return the oldest date of a first limit elements from set of
//   demodulators in the array demodulators. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

SysDate ClauseSetListGetMaxDate(ClauseSet_p *demodulators, int limit)
{
   int i;
   SysDate res = SysDateCreationTime();
   
   for(i=0; i<limit; i++)
   {
      assert(demodulators[i]);
      res = SysDateMaximum(res, demodulators[i]->date);
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetFind()
//
//   Given a clause and a clause set, try to find the clause in the
//   set. This is only useful for debugging, as usually clause should
//   know about the set it is in!
//
// Global Variables: -
//
// Side Effects    : - 
//
/----------------------------------------------------------------------*/

Clause_p ClauseSetFind(ClauseSet_p set, Clause_p clause)
{
   Clause_p handle, res = NULL;
   
   assert(set);
   assert(clause);

   for(handle = set->anchor->succ; handle!=set->anchor;
       handle=handle->succ)
   {
      if(handle == clause)
      {
	 res = handle;
	 assert(clause->set == set);
	 break;
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetFindById()
//
//   Given a clause ident and a clause set, try to find the clause in
//   the set. 
//
// Global Variables: -
//
// Side Effects    : - 
//
/----------------------------------------------------------------------*/

Clause_p ClauseSetFindById(ClauseSet_p set, long ident)
{
   Clause_p handle, res = NULL;
   
   assert(set);

   for(handle = set->anchor->succ; handle!=set->anchor;
       handle=handle->succ)
   {
      if(handle->ident == ident)
      {
	 res = handle;
	 assert(handle->set == set);
	 break;
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetRemoveEvaluations()
//
//   Remove all evaluations from the clauses in set.
//
// Global Variables: -
//
// Side Effects    : Memory operations...
//
/----------------------------------------------------------------------*/

void ClauseSetRemoveEvaluations(ClauseSet_p set)
{
   int i;
   Clause_p handle;
   for(i=0; i<set->eval_indices->size; i++)
   {
      EvalTreeFree(PDArrayElementP(set->eval_indices, i));
      PDArrayAssignP(set->eval_indices, i, NULL);
   }
   for(handle = set->anchor->succ; handle!=set->anchor;
       handle=handle->succ)
   {
      handle->evaluations = NULL;
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetFilterTautologies()
//
//   Given a clause set, remove all tautologies from it. Return number
//   of clauses removed.
//
// Global Variables: -
//
// Side Effects    : Changes set.
//
/----------------------------------------------------------------------*/

long ClauseSetFilterTautologies(ClauseSet_p set)
{
   Clause_p handle, next;
   long count = 0;

   assert(set);
   assert(!set->demod_index);
   
   handle = set->anchor->succ;
   while(handle != set->anchor)
   {      
      next = handle->succ;
      
      assert(handle);

      if(ClauseIsTrivial(handle))
      {
	 ClauseDetachParents(handle);	 
	 ClauseSetDeleteEntry(handle);
	 count++;
      }
      handle = next;
   }
   return count;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetDocumentProcessing()
//
//   If level >= 2, print a processing description for all the clauses
//   in the set.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClauseSetDocumentProcessing(FILE* out, long level, ClauseSet_p
				 set, char* operation)
{
   Clause_p handle;

   if(level>=2)
   {
      for(handle = set->anchor->succ; handle!=set->anchor; handle =
	     handle->succ)
      {
	 ClauseDocumentProcessing(out, level, handle, operation);
      }
   }   
}

#ifndef NDBUG

/*-----------------------------------------------------------------------
//
// Function: ClauseSetVerifyDemod()
//
//   Return true if pos->clause is in clause set, is a demodulator,
//   and if pos describes a potential maximal side in clause.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseSetVerifyDemod(ClauseSet_p demods, ClausePos_p pos)
{
   assert(demods);
   assert(pos);
   assert(TermPosIsTopPos(pos->pos));

   if(!ClauseSetFind(demods, pos->clause))
   {
      return false;
   }
   if(!ClauseIsDemodulator(pos->clause))
   {
      return false;
   }
   if((pos->side == RightSide) && EqnIsOriented(pos->clause->literals))
   {
      return false;
   }
   return true;      
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeVerifyIndex()
//
//   Check if all clauses in index are in demod as well.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool PDTreeVerifyIndex(PDTree_p root, ClauseSet_p demods)
{
   PStack_p stack = PStackAlloc(), trav;
   int              i;
   ClausePos_p      pos;
   PTree_p          entry;
   bool             res = true;
   
   if(root)
   {
      PStackPushP(stack, root);
      
      while(!PStackEmpty(stack))
      {
	 root = PStackPopP(stack);
	 
	 if(!root->demod_terms)
	 {
	    for(i=0; i<root->sig_size; i++)
	    {
	       if(root->f_alternatives[i])
	       {
		  PStackPushP(stack, root->f_alternatives[i]);
	       }
	    }
	    for(i=1; i<=root->max_var; i++)
	    {
	       if(PDArrayElementP(root->v_alternatives, i))
	       {
		  PStackPushP(stack, PDArrayElementP(root->v_alternatives, i));
	       }
	    }
	 }
	 else
	 {
	    trav = PTreeTraverseInit(root->demod_terms);
	    while((entry = PTreeTraverseNext(trav)))
	    {
	       pos = entry->key;
	       if(!ClauseSetFind(demods, pos->clause))
	       {
		  res = false;
		  /* printf("Fehler: %d\n", (int)pos->clause);
		  ClausePrint(stdout, pos->clause, true);
		  printf("\n"); */
	       }
	    }
	    PTreeTraverseExit(trav);
	    for(i=0; i<root->sig_size; i++)
	    {
	       assert(!root->f_alternatives[i]);
	    }
	 }
      }
   }
   PStackFree(stack);

   return res;
}

#endif

/*-----------------------------------------------------------------------
//
// Function: EqAxiomsPrint()
//
//   Print the equality axioms (symmetry, transitivity, refexivity,
//   substitutivity) for the given signature.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void EqAxiomsPrint(FILE* out, Sig_p sig, bool single_subst)
{
   FunCode i;
   int arity;

   if(TPTPFormatPrint)      
   {
      fprintf(out, 
	      "input_clause(eq_reflexive, axiom, [++equal(X,X)]).\n"
	      "input_clause(eq_symmetric, axiom,"
	      " [++equal(X,Y),--equal(Y,X)]).\n"
	      "input_clause(eq_transitive, axiom,"
	      " [++equal(X,Z),--equal(X,Y),--equal(Y,Z)]).\n");
      for(i=sig->internal_symbols+1; i<=sig->f_count; i++)
      {
	 if((arity=SigFindArity(sig, i)))
	 {
	    if(SigIsPredicate(sig,i))
	    {
	       tptp_eq_pred_axiom_print(out, SigFindName(sig,i),
					arity, single_subst);
	    }
	    else
	    {
	       tptp_eq_func_axiom_print(out, SigFindName(sig,i),
					arity, single_subst);
	    }
	 }
      }      
   }
   else
   {
      fprintf(out, 
	      "equal(X,X) <- .\n"
	      "equal(X,Y) <- equal(Y,X).\n"
	      "equal(X,Z) <- equal(X,Y), equal(Y,Z).\n");
      for(i=sig->internal_symbols+1; i<=sig->f_count; i++)
      {
	 if((arity=SigFindArity(sig, i)))
	 {
	    if(SigIsPredicate(sig,i))
	    {
	       eq_pred_axiom_print(out, SigFindName(sig,i), arity,
				   single_subst);
	    }
	    else
	    {
	       eq_func_axiom_print(out, SigFindName(sig,i), arity,
				   single_subst);
	    }
	 }
      }
   }
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetAddSymbolDistribution()
//
//   Count the occurrences of function symbols in set.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void ClauseSetAddSymbolDistribution(ClauseSet_p set, long *dist_array)
{
   Clause_p handle;
   
   for(handle = set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      ClauseAddSymbolDistribution(handle, dist_array);
   }
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


